from recruiter_app.elt_app.api import *
from recruiter_app.elt_app.collections import *
from recruiter_app.elt_app.service import *
from recruiter_app.utilities.helpers import mongo_global_init, get_prop
import asyncio
from tqdm import tqdm
import sys


class SOFactory(StackExchangeData):
    """
    Orchestrate actions between stackoverflow api calls and data update
    """
    user_counter = 1

    def __init__(self, mongo_host_type='local'):
        super().__init__()
        mongo_global_init(host_type=mongo_host_type)

    def get_tags_users_list(self, tags_threshold=5):
        """
        Fetch the users for which tags_top_qna_data_available
        is != True
        :param tags_threshold: threshold for number of top questions/answers
        :return:
        """
        print("Fetching existing users with tags_score")
        tags_users_dict = {}
        for data in StackOverflowUser.objects(tags_top_qna_data_available__ne=True):
            tags = []
            for tag in list(data.tags_score.keys()):
                tags_data_available = StackOverflowUser.objects().filter(
                    user_id=data.user_id)[0]['tags_data_available']
                if tags_data_available:
                    tags.append(data.tags_score[tag]['name'])
            tags_users_dict[data.user_id] = tags[:tags_threshold + 1]
        return tags_users_dict

    def get_incomplete_users(self, users_obj=None, has_more=True, threshold=10):
        """
        Fetch the users for which the tags_data_available is != True
        Also checks if users in users_obj have achieved the threshold
        :param users_obj: list of users for which the tags data is to be fetched
        :param has_more: tell the calling func that more users which do not have tags data are available
        :param threshold:
        :return: users_obj, more_users,
        reset_page - whenever a new user is added the page counter
        should reset to 0 again
        """
        reset_page = False
        if users_obj:
            for user_obj in users_obj:
                reset_page = mongo.service_update_tags_data_available(has_more,
                                                                      user_obj,
                                                                      threshold)
        users_obj = StackOverflowUser.objects(tags_data_available__ne=True).limit(100)
        more_users = True if len(StackOverflowUser.objects(
            tags_data_available__ne=True
        ).limit(101)) > 100 else False
        return users_obj, more_users, reset_page

    def update_user_tag_coll_qna(self):
        """
        Fetch the qna scores for every user for which tags_top_qna_data_available
        is != True
        Async issue :
        https://github.com/Big-Data-Programming/big-data-programming-april2022-team_apps/issues/3
        :return:
        """
        tag_users_data = self.get_tags_users_list()
        top_qna_results_list = []
        if tag_users_data:
            print(f"Updating top qna details for {tag_users_data} users")
            for user in tag_users_data:
                print(f"Getting QnA scores for user - {user}")
                target_tags = tag_users_data[user]
                for target_tag in target_tags:
                    top_qna_results_list.append((user, target_tag))
                    mongo.service_update_top_qna(user,
                                                 self.get_top_qna('question',
                                                                  user, target_tag),
                                                 target_tag,
                                                 'question')
                    mongo.service_update_top_qna(user,
                                                 self.get_top_qna('answer',
                                                                  user, target_tag),
                                                 target_tag,
                                                 'answer')
        else:
            print("No users document with tag details object, skipping top qna update")

    def create_user_tags_collection(self):
        """
        Fetch top tags of multiple users (max 100 users at a time)
        :return: None
        """
        print("Creating user tags details")
        existing_users_list, _, _ = self.get_incomplete_users()
        more_incomplete_users = True
        if existing_users_list:
            page = 1
            while more_incomplete_users:
                print(f"Looking in page {page}")
                for response_obj, has_more in self.get_tags(existing_users_list, page):
                    mongo.service_create_user_tags_records(response_obj)
                    existing_users_list, more_incomplete_users, reset_page = self.get_incomplete_users(
                        existing_users_list,
                        has_more)
                    page += 1
                    if reset_page:
                        print(f"New users added, resetting the page index")
                        page = 1
                    if not more_incomplete_users:
                        print("No more users are available for which tags are to be fetched")
                        break
                    self.update_user_tag_coll_qna()
        if not len(existing_users_list):
            """
            Just in case the top qna props were not updated for every 
            tag object
            """
            self.update_user_tag_coll_qna()
        print("Users Tag creation completed")

    def create_user_collection(self):
        """
        - Keep fetching the data until the mining target is achieved
          - Basic details of stackoverflow users and call mongodb's
            service_create_record
        """
        print("Creating user_details collection")
        target_counts = get_prop("MINING_TARGET").get("SO_COUNT")
        page_counts = target_counts // 100
        print(f"To achieve target {target_counts}, will look in {page_counts} pages")
        users_data_list = asyncio.run(self.get_users(page_counts=page_counts))
        for user_data in users_data_list:
            mongo.service_create_record(user_data)
        print("User collections created with basic details")
        self.create_user_tags_collection()


class GHFactory(GithubData):
    """
    Orchestrate actions between GitHub api calls and data update
    """
    # TODO : Feature to extract any github feature for existing candidate

    def __init__(self, mongo_host_type='local'):
        super().__init__()
        mongo_global_init(host_type=mongo_host_type)

    def build_users_collection(self, jd_name):
        """
        Using GitHub search api the users are searched with jd_name
        The search will keep happening till mining target is achieved
        or page till 10
        :param jd_name: name of the job description
        :return: None
        """
        if sys.platform.startswith("win"):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        target_counts = get_prop("MINING_TARGET").get("GH_COUNT")
        page = 0
        with tqdm(total=target_counts.get(jd_name)) as pbar:
            # GitHub search limit is 1000 results in total that's why page < 11
            while len(GithubUser.objects(for_jd=jd_name)) < target_counts.get(jd_name) and page < 11:
                print(f"\nGetting users in page {page}")
                data_json = self.get_users_list(jd_name, page)
                if len(data_json.get('items')) > 0:
                    user_list = [item.get('login') for item in data_json.get('items')]
                    user_list = asyncio.run(self.exec_get_user_foll(user_list))
                    print("\nFiltering out the users which are already there in collection")
                    user_list = list(filter(lambda x: not GithubUser.objects(login_name=x), user_list))
                    if user_list:
                        print(f"\nGetting the data for {len(user_list)} users")
                        results_users_obj, results_user_repo_obj = asyncio.run(self.exec_get_user_repo(user_list))
                        for user_details_obj, user_repo_obj in zip(results_users_obj, results_user_repo_obj):
                            if user_details_obj is not None and user_repo_obj is not None and len(user_repo_obj):
                                login_name = user_details_obj.get('login')
                                if login_name and not GithubUser.objects(login_name=login_name):
                                    print(f"Insert docs started for users in page {page}")
                                    mongo.service_create_github_users(jd_name,
                                                                      user_details_obj,
                                                                      user_repo_obj)
                    pbar.set_description("\nBuilding users collection progress ")
                    pbar.update(len(GithubUser.objects(for_jd=jd_name)))
                else:
                    print(f"No more users found for {jd_name}")
                    break
                page += 1


class RedditFactory(RedditData):
    """
        Orchestrate actions between reddit api calls and data update
    """

    def __init__(self, mongo_host_type='local'):
        super().__init__()
        mongo_global_init(host_type=mongo_host_type)

    def create_user_obj(self, content_objs, content_type='article'):
        """
        Formats the data so that the mongo service can update it in
        database
        :param content_objs:  dict of content data
        :param content_type: type of content - article or comment
        :return:
        """
        users_obj = {}
        article_ids = []
        after = None
        print(f"\nCreating users_obj from API response for {content_type}")
        for content_obj in tqdm(content_objs):
            content_obj = content_obj.get("data")
            if content_obj.get("author") not in users_obj:
                users_obj[content_obj.get("author")] = {}
            content_name = f"{content_type}_{content_obj.get('id')}"
            if content_name not in users_obj[content_obj.get("author")]:
                users_obj[content_obj.get("author")][content_name] = {}
            users_obj[content_obj.get("author")][content_name]["score"] = content_obj.get("score")
            users_obj[content_obj.get("author")][content_name]["awards"] = \
                content_obj.get("total_awards_received")
            if content_type == 'comment':
                users_obj[content_obj.get("author")][content_name]["body"] = \
                    content_obj.get("body")
            if content_type == 'article':
                users_obj[content_obj.get("author")][content_name]["title"] = content_obj.get("title")
                users_obj[content_obj.get("author")][content_name]["selftext"] = content_obj.get("selftext")
                users_obj[content_obj.get("author")][content_name]["upvote_ratio"] = \
                    content_obj.get("upvote_ratio")
                after = f"t3_{content_obj.get('id')}"
                article_ids.append(content_obj.get('id'))
        return users_obj, after, article_ids

    def create_users_from_comments(self, jd, article_ids):
        """
        Create users collection using data from comments
        :param jd: name of thread
        :param article_ids: list of article ids
        :return:
        """
        if sys.platform.startswith("win"):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        articles_comments_obj = asyncio.run(self.exec_get_article_comments(jd, article_ids))
        if articles_comments_obj:
            for article_comments_obj in articles_comments_obj:
                users_obj, _, _ = self.create_user_obj(article_comments_obj,
                                                       content_type="comment")
                mongo.service_create_rd_record(users_obj, jd)

    def create_user_from_articles(self, articles_obj, jd):
        """
        Create users collection from articles
        :param articles_obj:
        :param jd:
        :return:
        """
        users_obj, after, article_ids = self.create_user_obj(articles_obj,
                                                             content_type="article")
        mongo.service_create_rd_record(users_obj, jd)
        self.create_users_from_comments(jd, article_ids)
        return after

    def build_users_collection(self, jds):
        """
        Using reddit api the users are searched from specific threads
        defined in jds list the search will keep happening till mining
        target is achieved or page till 10
        :param jds: list of reddit threads
        :return: None
        """
        target_counts = get_prop("MINING_TARGET").get("REDDIT_COUNTS")
        for jd in jds:
            after = None
            while len(RedditUserProps.objects(for_jd=jd)) < target_counts.get(jd):
                articles_obj = self.get_articles(jd, after)
                if len(articles_obj) > 0:
                    last_article_id = self.create_user_from_articles(articles_obj, jd)
                    after = last_article_id
                    print(f"will start fetching results after {after}")
                    print(f"{target_counts.get(jd) - len(RedditUserProps.objects(for_jd=jd))} "
                          f"away from target for jd - {jd}")
                else:
                    print("No more articles found")
                    break
            print(f"{jd} completed moving on to next jd")




import mongoengine


class JobDescrConfig(mongoengine.Document):
    meta = {
        "db_alias": 'core',
        "collection": 'job_descr_config'
    }
    jd_name = mongoengine.StringField(required=True)
    skills_weightage = mongoengine.DictField()



import json
import os
import mongoengine
import requests
from tqdm import tqdm
import re
import pandas as pd


def read_passwd():
    """
    Reads the password from a environment variable.
    """
    pass_store = {}
    if os.environ.get('session_key') and os.environ.get('session_password'):
        pass_store['session_key'] = os.environ.get('session_key')
        pass_store['session_password'] = os.environ.get('session_password')
    else:
        raise Exception("Either of session_key or session_password not set in the environment")
    return pass_store


def get_prop(prop_name):
    """
    Returns the props in config/config.json.
    """
    with open("config/config.json", "r") as f:
        config = json.load(f)
    for key, value in config.items():
        if key == prop_name:
            return value
    raise Exception(f"prop {prop_name} not found in config file at : config/config.json")


def mongo_global_init(host_type='local'):
    """
    Initialises mongo db connection which is available globally for
    every document object, syntax for defining a DOM (document object mapper):
    Extend the class from mongoengine.Document and initialise below dict
    in a variable named meta
    meta = {
        "db_alias": 'core',
        "collection": 'candidate_ratings'
    }
    :param host_type: to use mongo cluster use - "cloud",
    to use the mongo local instance use - "local"
    default port for both is 27017
    :return: None
    """
    if host_type == 'local':
        alias_core = 'core'
        db_name = 'hush_recruiter'
        data = {"host": "localhost", "port": 27017}
        mongoengine.register_connection(alias=alias_core, name=db_name, **data)
    else:
        print(f'Host type is {host_type}')
        alias_core = 'core'
        db_name = 'hush_recruiter'
        username = os.environ['mongo_user']
        password = os.environ['mongo_pass']
        data = {"host": f"mongodb+srv://{username}:{password}@cluster0.oonus.mongodb.net/?retryWrites=true&w=majority",
                "port": 27017,
                "username": username,
                "password": password}
        mongoengine.register_connection(alias=alias_core, name=db_name, **data)


def get_auth_token_reddit():
    client_id = os.environ['reddit_client_id']
    client_secret = os.environ['reddit_client_secret']
    auth = requests.auth.HTTPBasicAuth(client_id, client_secret)
    reddit_username = os.environ['reddit_user_name']
    reddit_password = os.environ['reddit_password']
    data = {
        'grant_type': 'password',
        'username': reddit_username,
        'password': reddit_password
    }
    auth_url = get_prop("DATA_SOURCES").get("REDDIT").get("token_retrieval")
    headers = {'User-Agent': 'Tutorial2/0.0.1',
               'Content-Type': 'application/x-www-form-urlencoded'}
    res = requests.post(auth_url, auth=auth, data=data, headers=headers)
    if res.status_code == 200:
        print("Authorization successful")
        res = res.json()
        return f"Bearer {res.get('access_token')}", res.get("expires_in")
    raise Exception("Authorization with reddit was not successful")


def create_batch(users, batch_size=100, start=0,
                 description= "\nProcessing batch {0}"):
    if type(users) == int:
        users = [i for i in range(users)]
    if type(users) == list or type(users) == mongoengine.queryset.queryset.QuerySet:
        with tqdm(total=len(users)) as pbar:
            for i in range(start, len(users), batch_size):
                next_batch = [user for user in users[i:i + batch_size]]
                pbar.set_description(description.format(i))
                pbar.update(batch_size)
                yield next_batch


def get_ad_score_weights(data_obj, data_source):
    final_weight = None
    if data_source == "stackoverflow":
        final_weight = 5
        if data_obj.gold_badge_count != 0:
            final_weight -= 1
        if data_obj.silver_badge_count != 0:
            final_weight -= 1
        if data_obj.bronze_badge_count != 0:
            final_weight -= 1
        if data_obj.reputation != 0:
            final_weight -= 1
        if data_obj.accept_rate != 0:
            final_weight -= 1
    elif data_source == "github":
        final_weight = 3
        if data_obj.followers != 0:
            final_weight -= 1
        if data_obj.public_repos != 0:
            final_weight -= 1
        if data_obj.public_gists != 0:
            final_weight -= 1
    return final_weight


def get_base_label(label, base_map):
    for base_label in base_map:
        if label in base_map[base_label]:
            return base_label
    # TODO : Remove this return when all tags are labelled
    return label


def sort_by_id(str_list):
    if len(str_list) == 1:
        # handling due to data issue
        return str_list
    str_list.sort(key=lambda x: re.findall(r'\d+', x)[0], reverse=True)
    return str_list


def load_base_scores():
    return {"feature_question_score": 0,
            "feature_answer_score": 0,
            "feature_top_question": 0,
            "feature_top_answer": 0,
            "feature_top_a_accepted": 0,
            "feature_forks": 0,
            "feature_stars": 0,
            "feature_watchers": 0,
            "feature_repo_count": 0}


def get_scores_dict(keys_list, tag_obj):
    total_score = 0
    for key_name in keys_list:
        total_score += tag_obj.get(key_name)
    return total_score


def sort_by_scores(repo_list, lang_score_obj):
    sorted_list = [(repo, lang_score_obj.get(repo)) for repo in repo_list]
    sorted_list.sort(key=lambda x: x[1], reverse=True)
    return [e[0] for e in sorted_list]


def get_all_feature_scores(skill_obj):
    ft_scores_dict = {"so_q_score": skill_obj.get('feature_question_score'),
                      "so_a_score": skill_obj.get('feature_answer_score'),
                      "so_top_q_score": skill_obj.get('feature_top_question'),
                      "so_top_a_score": skill_obj.get('feature_top_answer'),
                      "so_top_a_isa_score": skill_obj.get('feature_top_a_accepted'),
                      "gh_forks_count": skill_obj.get('feature_forks'),
                      "gh_stars_count": skill_obj.get('feature_stars'),
                      "gh_watchers_count": skill_obj.get('feature_watchers'),
                      "gh_repo_count": skill_obj.get('feature_repo_count')}
    return ft_scores_dict


def get_so_tags_count(tags_prop_obj):
    skill_name_counts = {}
    len_tags_prop_obj = len(tags_prop_obj)
    for tag_idx in range(len_tags_prop_obj):
        for tag in tags_prop_obj[tag_idx].tags_score.keys():
            if tag.startswith('tag_'):
                tag_name = tags_prop_obj[tag_idx].tags_score.get(tag).get('name')
                if tag_name:
                    if tag_name not in skill_name_counts:
                        skill_name_counts[tag_name] = 0
                    skill_name_counts[tag_name] += 1
    df = pd.DataFrame().from_dict(skill_name_counts, orient='index')
    df.to_csv('so_skills_count.csv')


def get_gh_lang_count(tags_prop_obj):
    skill_name_counts = {}
    len_tags_prop_obj = len(tags_prop_obj)
    for tag_idx in range(len_tags_prop_obj):
        for tag in tags_prop_obj[tag_idx].lang_score.keys():
            if tag.startswith('lang_'):
                tag_name = tag.split("lang_")[1]
                if tag_name not in skill_name_counts:
                    skill_name_counts[tag_name] = 0
                skill_name_counts[tag_name] += 1
    df = pd.DataFrame().from_dict(skill_name_counts, orient='index')
    df.to_csv('gh_skills_count.csv')

# TODO : [Placeholder] for reddit additional score

# def get_rd_content_count(self):
#         tags_prop_obj = RedditUserProps.objects()
#         skill_name_counts = {}
#         len_tags_prop_obj = len(tags_prop_obj)
#         for tag_idx in range(len_tags_prop_obj):
#             for_jd = tags_prop_obj[tag_idx].for_jd
#             for tag in tags_prop_obj[tag_idx].thread_props.keys():
#                 if tag.startswith('comment_'):
#                     skill_name_counts[tags_prop_obj
#                     [tag_idx].thread_props[tag]['body']] = for_jd
#                 if tag.startswith('article_'):
#                     skill_name_counts[tags_prop_obj
#                     [tag_idx].thread_props[tag]['title']] = for_jd
#         df = pd.DataFrame().from_dict(skill_name_counts, orient='index')
#         df.to_csv('rd_skills_count.csv')


def get_skill_tag_map(df):
    """
    Dict structure : {"c#": ["c#", "c_sharp"]}
    :param df:
    :return:
    """
    tag_map = {}
    all_tags = set(df.tags.values)
    for tag in all_tags:
        if isinstance(tag, str):
            if tag not in tag_map:
                tag_map[tag] = []
            tag_map[tag].extend(df[df.tags == tag].tag_names.values)
    return tag_map


def run_prep_skill_count_csv(so_user_list, gh_user_list):
    print("Prep stackoverflow started")
    get_so_tags_count(so_user_list)
    print("Prep github started")
    get_gh_lang_count(gh_user_list)
    # TODO : [Placeholder] for reddit additional score
    # print("Prep reddit started")
    # get_rd_content_count(self.rd_user_list)



from flask import Flask, render_template, request
from flask_mail import Mail, Message
from itsdangerous import URLSafeTimedSerializer, SignatureExpired

app = Flask(__name__, template_folder='templates')
app.config['MAIL_SERVER'] = 'smtp.gmail.com'
app.config['MAIL_PORT'] = 465
app.config['MAIL_USERNAME'] = 'anamadheyaashoka003@gmail.com'
app.config['MAIL_PASSWORD'] = 'jrlcakojofqllwkz'
app.config['MAIL_USE_TLS'] = False
app.config['MAIL_USE_SSL'] = True
mail = Mail(app)

s = URLSafeTimedSerializer('Thisisasecret!')

#Routes to the host and send e-mail
@app.route("/")
def index():
    send_mail()
    return '<h1>Email sent!!</h1>'

# create a secure token using salt and sample text.
# send email to all the selected candidates with the test link 
def send_mail():
    sample_token = "sample_token"
    token = s.dumps(sample_token, salt='test_sample_token')
    msg = Message('Interview-Software Engineer', sender='anamadheyaashoka003@gmail.com',
                  recipients=['anamadheyaashoka005@gmail.com', 'anamadheyaashoka003@gmail.com'])
    msg.html = f"""
    <html><body>
    <p>Dear Candidates,</p>
    <p>To continue your interview process, please click the below link and submit your answers with in next 2 days: </p>
    <p><a href=`http://127.0.0.1:5000/questions/{token}`>Exam link</a></p>
    <p>Thanks and Regards,</p>
    <p>HushHush Recruiter HR</p>
    """
    mail.send(msg)
    print("Message send successfully")

# Route to the home page with set of questions 
@app.route("/home", methods=["post", "get"])
def home():
    sample_token = "test_sample_token"
    token=s.dumps(sample_token, salt='test_sample_token')
    print(f"Generated token: {token}")
    # form will be visible only when visible is set to true 
    visible = True
    if request.method == 'POST':
        candidate_name = request.form['candidate_name']
        question1 = request.form['question1']
        question2 = request.form['question2']
        question3 = request.form['question3']

        print(
            f"Name: {candidate_name}\n Question answered:Q1 answered: {question1}\n Q2 answered: {question2}\n Q3 "
            f"answered: {question3}")
        if question1 is not None or question2 is not None or question3 is not None:
            visible = False
    return render_template("home.html", visible=visible)

# check the expiration of the link which is sent in e-mail
@app.route('/questions/<token>')
def confirm_email(token):
    try:
        #setting link expire time to 100 seconds
        s.loads(token, salt='test_sample_token', max_age=100)
        return render_template("home.html", visible=True)
    except SignatureExpired:
        # display when link is expired
        return '<h1>Link is expired!</h1>'


if __name__ == "__main__":
    app.run(debug=True)



import os
from recruiter_app.utilities import helpers
import requests
import time
import asyncio
import aiohttp
from tqdm import tqdm


class GithubData:
    quota_remaining = None
    reset_on = 0
    quota_checked = False

    def __init__(self):
        config_props = helpers.get_prop("DATA_SOURCES").get("GITHUB")
        self.access_token = os.environ['github_pat']
        self.session = requests.sessions.Session()
        self.host = config_props.get('HOST')
        self.search_users = config_props.get('search_users')
        self.user_details = config_props.get('user_details')
        self.user_repos_list = config_props.get('user_repos_list')
        self.user_repo_details = config_props.get('user_repo_details')
        self.user_followers_details = config_props.get('user_followers_url')
        self.user_following_details = config_props.get('user_following_url')
        self.check_quota_url = config_props.get('check_quota')
        self.auth_headers = {"Authorization": os.environ['github_pat']}
        self.session.headers.update(self.auth_headers)
        self.task_users_obj = []
        self.task_user_repo_obj = []
        self.task_user_foll = []

    def _fetch(self, url, **kwargs):
        """
        Function to send the http request
        :param url: url
        :param kwargs: params in url
        :return: json response object
        """
        # time.sleep(2)
        url = "".join([self.host, url])
        content = self.session.get(url, **kwargs)
        if content.status_code != 200:
            raise Exception(f"HTTP request {url} failed, err code : {content.status_code} \n"
                            f"Message : {content.content}")
        quota_remaining = int(content.headers.get('X-RateLimit-Remaining'))
        reset_on = int(content.headers.get('X-RateLimit-Reset'))
        print(f"{'*' * 10}Quota remaining: {quota_remaining}{'*' * 10}")
        if quota_remaining == 0:
            print(f"Waiting for about {reset_on}s for github quota to reset ")
            time.sleep(int(reset_on) - time.time())
        return content.json()

    def get_users_list(self, query_keyword, page=0, page_limit=100):
        """
        Get the list of users
        :param query_keyword: search with job title
        :param page: page number
        :param page_limit: number of results per page
        :return: list of users
        """
        params = {
            "q": query_keyword,
            "page": page,
            "per_page": page_limit
        }
        url = self.search_users
        return self._fetch(url, params=params, headers=self.auth_headers)

    async def check_quota(self, session):
        url = "".join([self.host, self.check_quota_url])
        async with session.get(url, headers=self.auth_headers) as response:
            if response.status != 200:
                raise Exception(f"HTTP request {url} failed, err code : {response.status} \n"
                                f"Message : {response.content}")
            quota_remaining = int(response.headers.get('X-RateLimit-Remaining'))
            reset_on = int(response.headers.get('X-RateLimit-Reset'))
            print(f"{'*' * 10}Quota remaining: {quota_remaining}{'*' * 10}")
            return quota_remaining, reset_on

    async def _fetch_async(self, session, url, **kwargs):
        """
        Asyncio coroutine to send the http requests for multiple users
        Quota limit handling:
        GitHub apis are limited to 5k requests per hour, once this
        quota is exhausted the code will wait on reset_on parameter
        which is obtained from the api header response.
        All the async calls
        will wait till the quota is reset
        :param session: aiohttp session
        :param url: url
        :param kwargs: param in url
        :return: json response object
        """
        print(f"{'*' * 10}Quota remaining: {GithubData.quota_remaining}{'*' * 10}")
        complete_url = "".join([self.host, url])
        if GithubData.quota_remaining < 1:
            while GithubData.reset_on - time.time() > 0 and GithubData.quota_remaining < 1:
                print(f"Wait period is still in future so waiting for {GithubData.reset_on - time.time()}s")
                await asyncio.sleep(int(GithubData.reset_on - time.time()) + 2)
                print("Wait period is over, execution will continue now")
                if not GithubData.quota_checked:
                    print("Checking new quota now")
                    GithubData.quota_checked = True
                    GithubData.quota_remaining, GithubData.reset_on = await self.check_quota(session)
                    GithubData.quota_checked = False
            if GithubData.reset_on - time.time() < 0 and not GithubData.quota_checked:
                print("reset_on is old, updating reset_on and quota")
                GithubData.quota_checked = True
                GithubData.quota_remaining, GithubData.reset_on = await self.check_quota(session)
                GithubData.quota_checked = False
            print(f"New quota is {GithubData.quota_remaining}")
        await asyncio.sleep(10)
        try:
            GithubData.quota_remaining -= 1
            async with session.get(complete_url, **kwargs) as response:
                if response.status != 200:
                    if response.status == 403:
                        print(f"Cannot access this profile {complete_url}")
                    else:
                        raise Exception(f"HTTP request {url} failed, err code : {response.status} \n"
                                        f"Message : {response.content}")
                content = await response.json()
                return content
        except Exception as e:
            """
            This handling is done for 3types of errors I have seen so far and its root cause is unknown:
            - cannot connect to host TimeOutError is raised
            - cannot access the profile (same profile can be accessed manually)
            - 502 - bad gateway
            - 503 - resource unavailable
            """
            print(f"Connection failed with error : {e}")

    async def exec_get_user_repo(self, user_list, page_limit=100):
        """
        Get user details and user repositories details
        :param user_list:
        :param page_limit:
        :return: list of user details and list of user repo details
        """
        params = {
            "per_page": page_limit
        }
        async with aiohttp.ClientSession() as session:
            GithubData.quota_remaining, GithubData.reset_on = await self.check_quota(session)
            list_users_obj = []
            list_user_repo_obj = []
            for gh_user_batch in helpers.create_batch(user_list):
                for gh_user in gh_user_batch:
                    self.task_users_obj.append(self._fetch_async(session,
                                                                 self.user_details.format(gh_user),
                                                                 headers=self.auth_headers))
                    self.task_user_repo_obj.append(self._fetch_async(session,
                                                                     self.user_repos_list.format(gh_user),
                                                                     params=params,
                                                                     headers=self.auth_headers))
                print("Starting 100 async requests")
                results_users_obj = await asyncio.gather(*self.task_users_obj)
                list_users_obj.extend(results_users_obj)
                print("Wait for 10s before executing next 100 async requests")
                time.sleep(10)
                print("Starting 100 async requests")
                results_user_repo_obj = await asyncio.gather(*self.task_user_repo_obj)
                list_user_repo_obj.extend(results_user_repo_obj)
                self.task_users_obj = []
                self.task_user_repo_obj = []
        return list_users_obj, list_user_repo_obj

    async def exec_get_user_foll(self, user_list, page_limit=50):
        """
        Pick 50 followers + 50 following users
        :param user_list:
        :param page_limit:
        :return:
        """
        # TODO : Check if it is fine to include followers ?
        params = {
            "per_page": page_limit
        }
        async with aiohttp.ClientSession() as session:
            GithubData.quota_remaining, GithubData.reset_on = await self.check_quota(session)
            final_list = []
            for gh_user_batch in helpers.create_batch(user_list):
                for gh_user in gh_user_batch:
                    self.task_user_foll.append(self._fetch_async(session,
                                                                 self.user_followers_details.format(gh_user),
                                                                 params=params,
                                                                 headers=self.auth_headers))
                    self.task_user_foll.append(self._fetch_async(session,
                                                                 self.user_following_details.format(gh_user),
                                                                 params=params,
                                                                 headers=self.auth_headers))
                results_users_foll_obj = await asyncio.gather(*self.task_user_foll)
                self.task_user_foll = []
                final_users_list = []
                print("Preparing the final users list")
                if results_users_foll_obj:
                    for user_objs in results_users_foll_obj:
                        if user_objs and len(user_objs) > 0:
                            for user_obj in user_objs:
                                if user_obj:
                                    final_users_list.append(user_obj.get("login"))
                final_list.extend(final_users_list)
            print(f"Users list of size {len(final_list)} created")
        return final_list



from recruiter_app.utilities import helpers
import requests
import time
import asyncio
import aiohttp


class RedditData:
    global_wait = 0

    def __init__(self):
        self.session_token, self.session_time = helpers.get_auth_token_reddit()
        self.session = requests.sessions.Session()
        self.headers = {"Authorization": self.session_token}
        self.session.headers = self.headers

        self.host_url = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("HOST")
        self.web_dev_thread = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("web_dev_thread")
        self.devops_thread = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("devops_thread")
        self.cyber_seq_thread = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("cyber_seq_thread")
        self.ml_thread = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("ml_thread")
        self.py_thread = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("py_thread")
        self.comments = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("comments")
        self.user_details = helpers.get_prop("DATA_SOURCES").get("REDDIT").get("user_details")
        self.tasks = []

    def _fetch(self, url, **kwargs):
        # time.sleep(2)
        complete_url = "".join([self.host_url, url])
        content = self.session.get(complete_url, **kwargs)
        if content.status_code != 200:
            if content.status_code == 429:
                retry_count = 10
                while retry_count:
                    print("Too many requests exception returned from server , going to sleep 5s")
                    time.sleep(5)
                    content = self.session.get(complete_url, **kwargs)
                    if content.status_code == 200:
                        print("retry successful")
                        break
                    retry_count -= 1
                if retry_count == 0:
                    raise Exception("Retry failed, please rerun")
            else:
                raise Exception(f"HTTP request {url} failed, err code : {content.status_code} \n"
                                f"Message : {content.content}")
        quota_remaining = content.headers.get('x-ratelimit-remaining')
        reset_on = content.headers.get('x-ratelimit-reset')
        print(f"{'*' * 10}Quota remaining: {quota_remaining}{'*' * 10}")
        if quota_remaining == 0:
            print(f"Waiting for about {reset_on}s for github quota to reset ")
            time.sleep(int(reset_on) - time.time())
        return content

    def get_articles(self, thread_name, after=None, t='all', limit=100):
        """
        Get article details from the api
        :param thread_name: name of reddit thread
        :param after: show results after list index
        :param t: 'all' for all time top articles
        :param limit: number of articles per page
        :return: response object in json
        """
        url = f"{self.__getattribute__(thread_name)}/top"
        params = {
            't': t,
            "limit": limit,
        }
        if after:
            params['after'] = after
            params['count'] = 1
        res = self._fetch(url, params=params)
        res = res.json()
        res = res.get("data").get("children")
        return res

    async def get_article_comments(self, session, thread_name, article_id, limit=100):
        """
        Get article comments from the api
        :param session: aiohttp client session object
        :param thread_name: name of thread
        :param article_id: id of the article
        :param limit: number of comments per page
        :return: comment object in json
        """
        if RedditData.global_wait > 0:
            await asyncio.sleep(RedditData.global_wait)
            print(f"Resetting the global wait time of {RedditData.global_wait} to 0s")
            RedditData.global_wait = 0

        url = f"{self.__getattribute__(thread_name)}/comments/{article_id}"
        complete_url = "".join([self.host_url, url])
        params = {
            "limit": limit,
        }
        async with session.get(complete_url, params=params, headers=self.headers) as response:
            if response.status != 200:
                if response.status == 429:
                    retry_count = 10
                    while retry_count:
                        print("Too many requests exception returned from server , going to sleep 5s")
                        await asyncio.sleep(5)
                        response = await self.get_article_comments(session, thread_name,
                                                                   article_id, limit=100)
                        if response.status == 200:
                            print("retry successful")
                            break
                        retry_count -= 1
            quota_remaining = response.headers.get('x-ratelimit-remaining')
            reset_on = response.headers.get('x-ratelimit-reset')
            print(f"{'*' * 10}Quota remaining: {quota_remaining}{'*' * 10}")
            if quota_remaining == 0:
                print(f"Quota reached setting wait period of {reset_on}s")
                RedditData.global_wait = int(reset_on) - time.time()

            response = await response.json()
            if len(response) > 1:
                return response[1].get("data").get("children")
            print(f"There are no comments for article_id {article_id}")
            return None

    async def exec_get_article_comments(self, jd, article_ids):
        """
        Asyncio coroutine to fetch article comments
        :param jd: name of reddit thread
        :param article_ids: list of articles
        :return:
        """
        async with aiohttp.ClientSession() as session:
            for article_id in article_ids:
                self.tasks.append(self.get_article_comments(session, jd, article_id))
            results = await asyncio.gather(*self.tasks)
            self.tasks = []
        return results



import mongoengine


class StackOverflowUser(mongoengine.Document):
    meta = {
        "db_alias": 'core',
        "collection": 'so_users_details'
    }
    display_name = mongoengine.StringField(required=True)
    user_id = mongoengine.IntField(required=True, unique=True)
    gold_badge_count = mongoengine.IntField(default=0)
    silver_badge_count = mongoengine.IntField(default=0)
    bronze_badge_count = mongoengine.IntField(default=0)
    accept_rate = mongoengine.IntField(default=0)
    reputation = mongoengine.IntField(default=0)
    website_url = mongoengine.StringField()
    tags_data_available = mongoengine.BooleanField(default=False)
    tags_top_qna_data_available = mongoengine.BooleanField(default=False)
    tags_score = mongoengine.DictField()
    alter_ego = mongoengine.DictField()